@@ -69,6 +69,7 @@ public:
     OpenPGPFilter() : DefaultKeyFilter()
     {
         setIsOpenPGP(DefaultKeyFilter::Set);
+        setCanEncrypt(DefaultKeyFilter::Set);
     }
 };
 static std::shared_ptr<KeyFilter> s_pgpFilter = std::shared_ptr<KeyFilter> (new OpenPGPFilter);
@@ -78,6 +79,10 @@ public:
     OpenPGPSignFilter() : DefaultKeyFilter()
     {
         /* Also list unusable keys to make it transparent why they are unusable */
+        setDisabled(DefaultKeyFilter::NotSet);
+        setRevoked(DefaultKeyFilter::NotSet);
+        setExpired(DefaultKeyFilter::NotSet);
+        setCanSign(DefaultKeyFilter::Set);
         setHasSecret(DefaultKeyFilter::Set);
         setIsOpenPGP(DefaultKeyFilter::Set);
     }
@@ -89,6 +94,7 @@ public:
     SMIMEFilter(): DefaultKeyFilter()
     {
         setIsOpenPGP(DefaultKeyFilter::NotSet);
+        setCanEncrypt(DefaultKeyFilter::Set);
     }
 };
 static std::shared_ptr<KeyFilter> s_smimeFilter = std::shared_ptr<KeyFilter> (new SMIMEFilter);
@@ -97,6 +103,10 @@ class SMIMESignFilter: public DefaultKeyFilter
 public:
     SMIMESignFilter(): DefaultKeyFilter()
     {
+        setDisabled(DefaultKeyFilter::NotSet);
+        setRevoked(DefaultKeyFilter::NotSet);
+        setExpired(DefaultKeyFilter::NotSet);
+        setCanSign(DefaultKeyFilter::Set);
         setIsOpenPGP(DefaultKeyFilter::NotSet);
         setHasSecret(DefaultKeyFilter::Set);
     }
@@ -120,14 +130,15 @@ class ComboWidget: public QWidget
 public:
     explicit ComboWidget(KeySelectionCombo *combo):
         mCombo(combo),
-        mFilterBtn(new QPushButton)
+        mFilterBtn(new QPushButton),
+        mFromOverride(GpgME::UnknownProtocol)
     {
         auto hLay = new QHBoxLayout(this);
         hLay->addWidget(combo, 1);
         hLay->addWidget(mFilterBtn, 0);
 
         // Assume that combos start out with a filter
-        mFilterBtn->setIcon(QIcon::fromTheme("kt-remove-filters"));
+        mFilterBtn->setIcon(QIcon::fromTheme(QStringLiteral("kt-remove-filters")));
         mFilterBtn->setToolTip(i18n("Remove Filter"));
 
         // FIXME: This is ugly to enforce but otherwise the
@@ -140,11 +151,11 @@ public:
             if (curFilter.isEmpty()) {
                 mCombo->setIdFilter(mLastIdFilter);
                 mLastIdFilter = QString();
-                mFilterBtn->setIcon(QIcon::fromTheme("kt-remove-filters"));
+                mFilterBtn->setIcon(QIcon::fromTheme(QStringLiteral("kt-remove-filters")));
                 mFilterBtn->setToolTip(i18n("Remove Filter"));
             } else {
                 mLastIdFilter = curFilter;
-                mFilterBtn->setIcon(QIcon::fromTheme("kt-add-filters"));
+                mFilterBtn->setIcon(QIcon::fromTheme(QStringLiteral("kt-add-filters")));
                 mFilterBtn->setToolTip(i18n("Add Filter"));
                 mCombo->setIdFilter(QString());
             }
@@ -156,10 +167,21 @@ public:
         return mCombo;
     }
 
+    GpgME::Protocol fromOverride() const
+    {
+        return mFromOverride;
+    }
+
+    void setFromOverride(GpgME::Protocol proto)
+    {
+        mFromOverride = proto;
+    }
+
 private:
     KeySelectionCombo *mCombo;
     QPushButton *mFilterBtn;
     QString mLastIdFilter;
+    GpgME::Protocol mFromOverride;
 };
 
 static enum GpgME::UserID::Validity keyValidity(const GpgME::Key &key)
@@ -175,6 +197,17 @@ static enum GpgME::UserID::Validity keyValidity(const GpgME::Key &key)
 
     return validity;
 }
+
+static bool key_has_addr(const GpgME::Key &key, const QString &addr)
+{
+    for (const auto &uid: key.userIDs()) {
+        if (QString::fromStdString(uid.addrSpec()).toLower() == addr.toLower()) {
+            return true;
+        }
+    }
+    return false;
+}
+
 } // namespace
 
 class NewKeyApprovalDialog::Private
@@ -211,7 +244,7 @@ public:
         mScrollArea->setWidgetResizable(true);
         mScrollArea->setSizeAdjustPolicy(QAbstractScrollArea::AdjustToContentsOnFirstShow);
         mScrollArea->setFrameStyle(QFrame::NoFrame);
-        mScrollLayout->setMargin(0);
+        mScrollLayout->setContentsMargins(0, 0, 0, 0);
 
         q->setWindowTitle(i18n("Security approval"));
 
@@ -272,8 +305,10 @@ public:
     {
         const auto &addr = combo->property("address").toString();
         auto job = new QGpgME::DefaultKeyGenerationJob(q);
-        auto progress = new Kleo::ProgressDialog(job, i18n("Generating key for '%1'...", addr),
-                                                 q);
+        auto progress = new Kleo::ProgressDialog(job, i18n("Generating key for '%1'...", addr) + QStringLiteral("\n\n") +
+                                                 i18n("This can take several minutes."), q);
+        progress->setWindowFlags(progress->windowFlags() & ~Qt::WindowContextHelpButtonHint);
+        progress->setWindowTitle(i18n("Key generation"));
         progress->setModal(true);
         progress->setAutoClose(true);
         progress->setMinimumDuration(0);
@@ -292,7 +327,7 @@ public:
     {
         mLastError = result.error();
         if (!mLastError || mLastError.isCanceled()) {
-            combo->setDefaultKey(result.fingerprint());
+            combo->setDefaultKey(QString::fromLatin1(result.fingerprint()), GpgME::OpenPGP);
             connect (combo, &KeySelectionCombo::keyListingFinished, q, [this, job] () {
                     mRunningJobs.removeAll(job);
                 });
@@ -305,13 +340,56 @@ public:
     void checkAccepted()
     {
         if (mLastError || mLastError.isCanceled()) {
-            KMessageBox::error(q, mLastError.asString(), i18n("Operation Failed"));
+            KMessageBox::error(q, QString::fromLocal8Bit(mLastError.asString()), i18n("Operation Failed"));
             mRunningJobs.clear();
             return;
         }
-        if (mRunningJobs.empty()) {
-            q->accept();
+
+        if (!mRunningJobs.empty()) {
+            return;
         }
+        /* Save the keys */
+        bool isPGP = mFormatBtns->checkedId() == 1;
+        bool isSMIME = mFormatBtns->checkedId() == 2;
+
+        mAcceptedEnc.clear();
+        mAcceptedSig.clear();
+
+        for (const auto combo: mEncCombos) {
+            const auto &addr = combo->property("address").toString();
+            const auto &key = combo->currentKey();
+            if (!combo->isVisible()) {
+                continue;
+            }
+            if (isSMIME && key.protocol() != GpgME::CMS) {
+                continue;
+            }
+            if (isPGP && key.protocol() != GpgME::OpenPGP) {
+                continue;
+            }
+            if (mAcceptedEnc.contains(addr)) {
+                mAcceptedEnc[addr].push_back(key);
+            } else {
+                std::vector<GpgME::Key> vec;
+                vec.push_back(key);
+                mAcceptedEnc.insert(addr, vec);
+            }
+        }
+        for (const auto combo: mSigningCombos) {
+            const auto key = combo->currentKey();
+            if (!combo->isVisible()) {
+                continue;
+            }
+            if (isSMIME && key.protocol() != GpgME::CMS) {
+                continue;
+            }
+            if (isPGP && key.protocol() != GpgME::OpenPGP) {
+                continue;
+            }
+            mAcceptedSig.push_back(combo->currentKey());
+        }
+
+        q->accept();
     }
 
     void accepted()
@@ -347,9 +425,26 @@ public:
         }
         for (auto combo: mSigningCombos) {
             combo->setKeyFilter(mCurSigFilter);
+            auto widget = qobject_cast <ComboWidget *>(combo->parentWidget());
+            if (!widget) {
+                qCDebug(LIBKLEO_LOG) << "Failed to find signature combo widget";
+                continue;
+            }
+            widget->setVisible(widget->fromOverride() == GpgME::UnknownProtocol ||
+                               ((isSMIME && widget->fromOverride() == GpgME::CMS) ||
+                                (isPGP && widget->fromOverride() == GpgME::OpenPGP)));
         }
         for (auto combo: mEncCombos) {
             combo->setKeyFilter(mCurEncFilter);
+
+            auto widget = qobject_cast <ComboWidget *>(combo->parentWidget());
+            if (!widget) {
+                qCDebug(LIBKLEO_LOG) << "Failed to find combo widget";
+                continue;
+            }
+            widget->setVisible(widget->fromOverride() == GpgME::UnknownProtocol ||
+                               ((isSMIME && widget->fromOverride() == GpgME::CMS) ||
+                                (isPGP && widget->fromOverride() == GpgME::OpenPGP)));
         }
     }
 
@@ -358,7 +453,7 @@ public:
         auto combo = new KeySelectionCombo();
         combo->setKeyFilter(mCurSigFilter);
         if (!key.isNull()) {
-            combo->setDefaultKey(key.primaryFingerprint());
+            combo->setDefaultKey(QString::fromLatin1(key.primaryFingerprint()), key.protocol());
         }
         if (key.isNull() && mProto != GpgME::CMS) {
             combo->appendCustomItem(QIcon::fromTheme(QStringLiteral("document-new")),
@@ -374,6 +469,9 @@ public:
         connect(combo, &KeySelectionCombo::currentKeyChanged, q, [this] () {
             updateOkButton();
         });
+        connect(combo, QOverload<int>::of(&QComboBox::currentIndexChanged), q, [this] () {
+            updateOkButton();
+        });
 
         return new ComboWidget(combo);
     }
@@ -395,9 +493,12 @@ public:
             for (const auto &key: resolved[addr])
             {
                 auto comboWidget = createSigningCombo(addr, key);
-                if (resolved[addr].size() == 1) {
+                if (key_has_addr (key, addr)) {
                     comboWidget->combo()->setIdFilter(addr);
                 }
+                if (resolved[addr].size() > 1) {
+                    comboWidget->setFromOverride(key.protocol());
+                }
                 sigLayout->addWidget(comboWidget);
             }
         }
@@ -425,7 +526,8 @@ public:
             auto combo = new KeySelectionCombo(false);
             combo->setKeyFilter(mCurEncFilter);
             if (!key.isNull()) {
-                combo->setDefaultKey(key.primaryFingerprint());
+                combo->setDefaultKey(QString::fromLatin1(key.primaryFingerprint()),
+                                     key.protocol());
             }
 
             if (mSender == addr && key.isNull()) {
@@ -436,18 +538,24 @@ public:
             combo->appendCustomItem(QIcon::fromTheme(QStringLiteral("emblem-unavailable")),
                     i18n("Ignore recipient"), IgnoreKey);
 
-            if (keys.size () == 1) {
+            if (key.isNull() || key_has_addr (key, addr)) {
                 combo->setIdFilter(addr);
             }
 
             connect(combo, &KeySelectionCombo::currentKeyChanged, q, [this] () {
-                    updateOkButton();
+                updateOkButton();
+            });
+            connect(combo, QOverload<int>::of(&QComboBox::currentIndexChanged), q, [this] () {
+                updateOkButton();
             });
 
             mEncCombos << combo;
             mAllCombos << combo;
             combo->setProperty("address", addr);
             auto comboWidget = new ComboWidget(combo);
+            if (keys.size() > 1) {
+                comboWidget->setFromOverride(key.protocol());
+            }
             encGrid->addWidget(comboWidget, encGrid->rowCount(), 0, 1, 2);
         }
     }
@@ -481,6 +589,7 @@ public:
     {
         static QString origOkText = mOkButton->text();
         bool isGenerate = false;
+        bool isAllIgnored = true;
         // Check if generate is selected.
         for (auto combo: mAllCombos) {
             auto act = combo->currentData(Qt::UserRole).toInt();
@@ -488,7 +597,13 @@ public:
                 mOkButton->setText(i18n("Generate"));
                 isGenerate = true;
             }
+            if (act != IgnoreKey) {
+                isAllIgnored = false;
+            }
         }
+
+        mOkButton->setEnabled(!isAllIgnored);
+
         if (!isGenerate) {
             mOkButton->setText(origOkText);
         }
@@ -520,8 +635,8 @@ public:
         }
 
         mOkButton->setIcon(QIcon::fromTheme(de_vs
-                    ? "security-high"
-                    : "security-medium"));
+                    ? QStringLiteral("security-high")
+                    : QStringLiteral("security-medium")));
         mOkButton->setStyleSheet(QStringLiteral("background-color: ") + (de_vs
                     ? QStringLiteral("#D5FAE2")  // KColorScheme(QPalette::Active, KColorScheme::View).background(KColorScheme::PositiveBackground).color().name()
                     : QStringLiteral("#FAE9EB"))); //KColorScheme(QPalette::Active, KColorScheme::View).background(KColorScheme::NegativeBackground).color().name()));
@@ -565,6 +680,8 @@ public:
     QList <QGpgME::Job *> mRunningJobs;
     GpgME::Error mLastError;
     QLabel *mComplianceLbl;
+    QMap<QString, std::vector<GpgME::Key> > mAcceptedEnc;
+    std::vector<GpgME::Key> mAcceptedSig;
 };
 
 NewKeyApprovalDialog::NewKeyApprovalDialog(const QMap<QString, std::vector<GpgME::Key> > &resolvedSigningKeys,
@@ -581,6 +698,7 @@ NewKeyApprovalDialog::NewKeyApprovalDialog(const QMap<QString, std::vector<GpgME
 {
     d->addSigningKeys(resolvedSigningKeys, unresolvedSigKeys);
     d->addEncryptionKeys(resolvedRecp, unresolvedRecp);
+    d->updateFilter();
     d->updateOkButton();
 
     const auto size = sizeHint();
@@ -590,30 +708,12 @@ NewKeyApprovalDialog::NewKeyApprovalDialog(const QMap<QString, std::vector<GpgME
 
 std::vector<GpgME::Key> NewKeyApprovalDialog::signingKeys()
 {
-    std::vector <GpgME::Key> ret;
-
-    for (const auto combo: d->mSigningCombos) {
-        ret.push_back(combo->currentKey());
-    }
-
-    return ret;
+    return d->mAcceptedSig;
 }
 
 QMap <QString, std::vector<GpgME::Key> > NewKeyApprovalDialog::encryptionKeys()
 {
-    QMap <QString, std::vector<GpgME::Key> > ret;
-    for (const auto combo: d->mEncCombos) {
-        const auto &addr = combo->property("address").toString();
-        const auto &key = combo->currentKey();
-        if (ret.contains(addr)) {
-            ret[addr].push_back(key);
-        } else {
-            std::vector<GpgME::Key> vec;
-            vec.push_back(key);
-            ret.insert(addr, vec);
-        }
-    }
-    return ret;
+    return d->mAcceptedEnc;
 }
 
 #include "newkeyapprovaldialog.moc"
